from ._constrained_control_demo_1 import _constrained_control_demo_1
from ._control_demo_1 import _control_demo_1



class Demo:
    
    @staticmethod
    def constrained_control_demo_1():
        return _constrained_control_demo_1()
    
    @staticmethod
    def control_demo_1():
        return _control_demo_1()
